from shapely.geometry import Polygon, MultiPolygon, LineString
import pytest
from .shapely import polyfill


@pytest.fixture
def polygon():
    return Polygon([(48, 18), (49, 18), (49, 19), (48, 19)])


@pytest.fixture
def polygon_b():
    return Polygon([(54, 11), (56, 11), (56, 12), (54, 12)])


@pytest.fixture
def polygon_with_hole():
    return Polygon(
        [(48, 18), (49, 18), (49, 19), (48, 19)],
        [[(48.4, 18.2), (48.8, 18.2), (48.8, 18.6), (48.4, 18.6)]],
    )


@pytest.fixture
def multipolygon(polygon, polygon_b):
    return MultiPolygon([polygon, polygon_b])


@pytest.fixture
def line():
    return LineString([(0, 0), (1, 0), (1, 1)])


def test_polyfill_polygon(polygon):
    expected = set(["811e3ffffffffff"])
    result = polyfill(polygon, 1)
    assert expected == result


def test_polyfill_multipolygon(multipolygon):
    expected = set(["811e3ffffffffff", "811f3ffffffffff"])
    result = polyfill(multipolygon, 1)
    assert expected == result


def test_polyfill_polygon_with_hole(polygon_with_hole):
    expected = set()
    result = polyfill(polygon_with_hole, 1)
    assert expected == result


def test_polyfill_wrong_type(line):
    with pytest.raises(TypeError, match=".*Unknown type.*"):
        polyfill(line, 1)



from h3 import h3
import pytest

from .decorator import catch_invalid_h3_address


def test_catch_invalid_h3_address():
    @catch_invalid_h3_address
    def safe_h3_to_parent(h3_address):
        return h3.h3_to_parent(h3_address, 1)

    with pytest.raises(ValueError):
        safe_h3_to_parent("a")  # Originally ValueError

    with pytest.raises(ValueError):
        safe_h3_to_parent(1)  # Originally TypeError

    with pytest.raises(ValueError):
        safe_h3_to_parent("891f1d48177fff1")  # Originally H3CellError



from . import h3pandas  # noqa: F401
from h3 import h3
import pytest
from shapely.geometry import Polygon, box
import pandas as pd
import geopandas as gpd
from geopandas.testing import assert_geodataframe_equal


# TODO: Make sure methods are tested both for
#  DataFrame and GeoDataFrame (where applicable)
# TODO: Test return_geometry functionality

# Fixtures


@pytest.fixture
def basic_dataframe():
    """DataFrame with lat and lng columns"""
    return pd.DataFrame({"lat": [50, 51], "lng": [14, 15]})


@pytest.fixture
def basic_geodataframe(basic_dataframe):
    """GeoDataFrame with POINT geometry"""
    geometry = gpd.points_from_xy(basic_dataframe["lng"], basic_dataframe["lat"])
    return gpd.GeoDataFrame(geometry=geometry, crs="epsg:4326")


@pytest.fixture
def basic_geodataframe_polygon(basic_geodataframe):
    geom = box(0, 0, 1, 1)
    return gpd.GeoDataFrame(geometry=[geom], crs="epsg:4326")


@pytest.fixture
def basic_geodataframe_polygons(basic_geodataframe):
    geoms = [box(0, 0, 1, 1), box(0, 0, 2, 2)]
    return gpd.GeoDataFrame(geometry=geoms, crs="epsg:4326")


@pytest.fixture
def basic_dataframe_with_values(basic_dataframe):
    """DataFrame with lat and lng columns and values"""
    return basic_dataframe.assign(val=[2, 5])


@pytest.fixture
def basic_geodataframe_with_values(basic_geodataframe):
    """GeoDataFrame with POINT geometry and values"""
    return basic_geodataframe.assign(val=[2, 5])


@pytest.fixture
def indexed_dataframe(basic_dataframe):
    """DataFrame with lat, lng and resolution 9 H3 index"""
    return basic_dataframe.assign(
        h3_09=["891e3097383ffff", "891e2659c2fffff"]
    ).set_index("h3_09")


@pytest.fixture
def h3_dataframe_with_values():
    """DataFrame with resolution 9 H3 index and values"""
    index = ["891f1d48177ffff", "891f1d48167ffff", "891f1d4810fffff"]
    return pd.DataFrame({"val": [1, 2, 5]}, index=index)


@pytest.fixture
def h3_geodataframe_with_values(h3_dataframe_with_values):
    """GeoDataFrame with resolution 9 H3 index, values, and Hexagon geometries"""
    geometry = [
        Polygon(h3.h3_to_geo_boundary(h, True)) for h in h3_dataframe_with_values.index
    ]
    return gpd.GeoDataFrame(
        h3_dataframe_with_values, geometry=geometry, crs="epsg:4326"
    )


# Tests: H3 API


def test_geo_to_h3(basic_dataframe):
    result = basic_dataframe.h3.geo_to_h3(9)
    expected = basic_dataframe.assign(
        h3_09=["891e3097383ffff", "891e2659c2fffff"]
    ).set_index("h3_09")

    pd.testing.assert_frame_equal(expected, result)


def test_geo_to_h3_geo(basic_geodataframe):
    result = basic_geodataframe.h3.geo_to_h3(9)
    expected = basic_geodataframe.assign(
        h3_09=["891e3097383ffff", "891e2659c2fffff"]
    ).set_index("h3_09")

    pd.testing.assert_frame_equal(expected, result)


def test_geo_to_h3_polygon(basic_geodataframe_polygon):
    with pytest.raises(ValueError):
        basic_geodataframe_polygon.h3.geo_to_h3(9)


def test_h3_to_geo(indexed_dataframe):
    lats = [50.000551554902586, 51.000121447274736]
    lngs = [14.000372151097624, 14.999768926738376]
    geometry = gpd.points_from_xy(x=lngs, y=lats, crs="epsg:4326")
    expected = gpd.GeoDataFrame(indexed_dataframe, geometry=geometry)
    result = indexed_dataframe.h3.h3_to_geo()
    assert_geodataframe_equal(expected, result, check_less_precise=True)


def test_h3_to_geo_boundary(indexed_dataframe):
    h1 = (
        (13.997875502962215, 50.00126530465277),
        (13.997981974191347, 49.99956539765703),
        (14.000478563108897, 49.99885162163456),
        (14.002868770645003, 49.99983773856239),
        (14.002762412857178, 50.00153765760209),
        (14.000265734090084, 50.00225144767143),
        (13.997875502962215, 50.00126530465277),
    )
    h2 = (
        (14.9972390328545, 51.00084372147122),
        (14.99732334029277, 50.99916437137475),
        (14.999853173220332, 50.99844207137708),
        (15.002298787294139, 50.99939910547163),
        (15.002214597747209, 51.00107846572982),
        (14.999684676233445, 51.00180078173323),
        (14.9972390328545, 51.00084372147122),
    )
    geometry = [Polygon(h1), Polygon(h2)]

    result = indexed_dataframe.h3.h3_to_geo_boundary()
    expected = gpd.GeoDataFrame(indexed_dataframe, geometry=geometry, crs="epsg:4326")
    assert_geodataframe_equal(expected, result, check_less_precise=True)


def test_h3_to_geo_boundary_wrong_index(indexed_dataframe):
    indexed_dataframe.index = [str(indexed_dataframe.index[0])] + ["invalid"]
    with pytest.raises(ValueError):
        indexed_dataframe.h3.h3_to_geo_boundary()


def test_h3_to_parent(h3_dataframe_with_values):
    h3_parent = "811f3ffffffffff"
    result = h3_dataframe_with_values.h3.h3_to_parent(1)
    expected = h3_dataframe_with_values.assign(h3_01=h3_parent)

    pd.testing.assert_frame_equal(expected, result)


def test_h3_to_center_child(indexed_dataframe):
    expected = indexed_dataframe.assign(
        h3_center_child=["8a1e30973807fff", "8a1e2659c2c7fff"]
    )
    result = indexed_dataframe.h3.h3_to_center_child()
    pd.testing.assert_frame_equal(expected, result)


def test_empty_polyfill(h3_geodataframe_with_values):
    expected = h3_geodataframe_with_values.assign(h3_polyfill=[list(), list(), list()])
    result = h3_geodataframe_with_values.h3.polyfill(1)
    assert_geodataframe_equal(expected, result)


def test_polyfill(h3_geodataframe_with_values):
    expected_cells = [
        {
            "8a1f1d481747fff",
            "8a1f1d48174ffff",
            "8a1f1d481757fff",
            "8a1f1d48175ffff",
            "8a1f1d481767fff",
            "8a1f1d48176ffff",
            "8a1f1d481777fff",
        },
        {
            "8a1f1d481647fff",
            "8a1f1d48164ffff",
            "8a1f1d481657fff",
            "8a1f1d48165ffff",
            "8a1f1d481667fff",
            "8a1f1d48166ffff",
            "8a1f1d481677fff",
        },
        {
            "8a1f1d4810c7fff",
            "8a1f1d4810cffff",
            "8a1f1d4810d7fff",
            "8a1f1d4810dffff",
            "8a1f1d4810e7fff",
            "8a1f1d4810effff",
            "8a1f1d4810f7fff",
        },
    ]
    expected = h3_geodataframe_with_values.assign(h3_polyfill=expected_cells)
    result = h3_geodataframe_with_values.h3.polyfill(10)
    result["h3_polyfill"] = result["h3_polyfill"].apply(
        set
    )  # Convert to set for testing
    assert_geodataframe_equal(expected, result)


def test_polyfill_explode(h3_geodataframe_with_values):
    expected_indices = set().union(
        *[
            {
                "8a1f1d481747fff",
                "8a1f1d48174ffff",
                "8a1f1d481757fff",
                "8a1f1d48175ffff",
                "8a1f1d481767fff",
                "8a1f1d48176ffff",
                "8a1f1d481777fff",
            },
            {
                "8a1f1d481647fff",
                "8a1f1d48164ffff",
                "8a1f1d481657fff",
                "8a1f1d48165ffff",
                "8a1f1d481667fff",
                "8a1f1d48166ffff",
                "8a1f1d481677fff",
            },
            {
                "8a1f1d4810c7fff",
                "8a1f1d4810cffff",
                "8a1f1d4810d7fff",
                "8a1f1d4810dffff",
                "8a1f1d4810e7fff",
                "8a1f1d4810effff",
                "8a1f1d4810f7fff",
            },
        ]
    )
    result = h3_geodataframe_with_values.h3.polyfill(10, explode=True)
    assert len(result) == len(h3_geodataframe_with_values) * 7
    assert set(result["h3_polyfill"]) == expected_indices
    assert not result["val"].isna().any()


def test_polyfill_explode_unequal_lengths(basic_geodataframe_polygons):
    expected_indices = {
        "83754efffffffff",
        "83756afffffffff",
        "83754efffffffff",
        "837541fffffffff",
        "83754cfffffffff",
    }
    result = basic_geodataframe_polygons.h3.polyfill(3, explode=True)
    assert len(result) == 5
    assert set(result["h3_polyfill"]) == expected_indices


def test_cell_area(indexed_dataframe):
    expected = indexed_dataframe.assign(
        h3_cell_area=[0.09937867173389912, 0.09775508251476996]
    )
    result = indexed_dataframe.h3.cell_area()
    pd.testing.assert_frame_equal(expected, result)


def test_h3_get_resolution(h3_dataframe_with_values):
    expected = h3_dataframe_with_values.assign(h3_resolution=9)
    result = h3_dataframe_with_values.h3.h3_get_resolution()
    pd.testing.assert_frame_equal(expected, result)


def test_h3_get_base_cell(indexed_dataframe):
    expected = indexed_dataframe.assign(h3_base_cell=[15, 15])
    result = indexed_dataframe.h3.h3_get_base_cell()
    pd.testing.assert_frame_equal(expected, result)


def test_h3_0_ring(indexed_dataframe):
    expected = indexed_dataframe.assign(
        h3_k_ring=[[h] for h in indexed_dataframe.index]
    )
    result = indexed_dataframe.h3.k_ring(0)
    pd.testing.assert_frame_equal(expected, result)


def test_h3_k_ring(indexed_dataframe):
    expected_indices = [
        {
            "891e3097383ffff",
            "891e3097387ffff",
            "891e309738bffff",
            "891e309738fffff",
            "891e3097393ffff",
            "891e3097397ffff",
            "891e309739bffff",
        },
        {
            "891e2659893ffff",
            "891e2659897ffff",
            "891e2659c23ffff",
            "891e2659c27ffff",
            "891e2659c2bffff",
            "891e2659c2fffff",
            "891e2659d5bffff",
        },
    ]
    expected = indexed_dataframe.assign(h3_k_ring=expected_indices)
    result = indexed_dataframe.h3.k_ring()
    result["h3_k_ring"] = result["h3_k_ring"].apply(set)  # Convert to set for testing
    pd.testing.assert_frame_equal(expected, result)


def test_h3_k_ring_explode(indexed_dataframe):
    expected_indices = set().union(
        *[
            {
                "891e3097383ffff",
                "891e3097387ffff",
                "891e309738bffff",
                "891e309738fffff",
                "891e3097393ffff",
                "891e3097397ffff",
                "891e309739bffff",
            },
            {
                "891e2659893ffff",
                "891e2659897ffff",
                "891e2659c23ffff",
                "891e2659c27ffff",
                "891e2659c2bffff",
                "891e2659c2fffff",
                "891e2659d5bffff",
            },
        ]
    )
    result = indexed_dataframe.h3.k_ring(explode=True)
    assert len(result) == len(indexed_dataframe) * 7
    assert set(result["h3_k_ring"]) == expected_indices
    assert not result["lat"].isna().any()


def test_h3_0_hex_ring(indexed_dataframe):
    expected = indexed_dataframe.assign(
        h3_hex_ring=[[h] for h in indexed_dataframe.index]
    )
    result = indexed_dataframe.h3.hex_ring(0)
    pd.testing.assert_frame_equal(expected, result)


def test_h3_0_hex_ring_explode(indexed_dataframe):
    expected = indexed_dataframe.assign(
        h3_hex_ring=[h for h in indexed_dataframe.index]
    )
    result = indexed_dataframe.h3.hex_ring(0, True)
    pd.testing.assert_frame_equal(expected, result)


def test_h3_hex_ring(indexed_dataframe):
    expected_indices = [
        {
            "891e3097387ffff",
            "891e309738bffff",
            "891e309738fffff",
            "891e3097393ffff",
            "891e3097397ffff",
            "891e309739bffff",
        },
        {
            "891e2659893ffff",
            "891e2659897ffff",
            "891e2659c23ffff",
            "891e2659c27ffff",
            "891e2659c2bffff",
            "891e2659d5bffff",
        },
    ]
    expected = indexed_dataframe.assign(h3_hex_ring=expected_indices)
    result = indexed_dataframe.h3.hex_ring()
    result["h3_hex_ring"] = result["h3_hex_ring"].apply(
        set
    )  # Convert to set for testing
    pd.testing.assert_frame_equal(expected, result)


def test_h3_hex_ring_explode(indexed_dataframe):
    expected_indices = set().union(
        *[
            {
                "891e3097387ffff",
                "891e309738bffff",
                "891e309738fffff",
                "891e3097393ffff",
                "891e3097397ffff",
                "891e309739bffff",
            },
            {
                "891e2659893ffff",
                "891e2659897ffff",
                "891e2659c23ffff",
                "891e2659c27ffff",
                "891e2659c2bffff",
                "891e2659d5bffff",
            },
        ]
    )
    result = indexed_dataframe.h3.hex_ring(explode=True)
    assert len(result) == len(indexed_dataframe) * 6
    assert set(result["h3_hex_ring"]) == expected_indices
    assert not result["lat"].isna().any()


def test_h3_is_valid(indexed_dataframe):
    indexed_dataframe.index = [str(indexed_dataframe.index[0])] + ["invalid"]
    expected = indexed_dataframe.assign(h3_is_valid=[True, False])
    result = indexed_dataframe.h3.h3_is_valid()
    pd.testing.assert_frame_equal(expected, result)


def test_h3_get_resolution_index_only(h3_dataframe_with_values):
    del h3_dataframe_with_values["val"]
    expected = h3_dataframe_with_values.assign(h3_resolution=9)
    result = h3_dataframe_with_values.h3.h3_get_resolution()
    pd.testing.assert_frame_equal(expected, result)


# Tests: Aggregate functions


def test_geo_to_h3_aggregate(basic_dataframe_with_values):
    result = basic_dataframe_with_values.h3.geo_to_h3_aggregate(
        1, return_geometry=False
    )
    expected = pd.DataFrame({"h3_01": ["811e3ffffffffff"], "val": [2 + 5]}).set_index(
        "h3_01"
    )

    pd.testing.assert_frame_equal(expected, result)


def test_geo_to_h3_aggregate_geo(basic_geodataframe_with_values):
    result = basic_geodataframe_with_values.h3.geo_to_h3_aggregate(
        1, return_geometry=False
    )
    expected = pd.DataFrame({"h3_01": ["811e3ffffffffff"], "val": [2 + 5]}).set_index(
        "h3_01"
    )

    pd.testing.assert_frame_equal(expected, result)


def test_h3_to_parent_aggregate(h3_geodataframe_with_values):
    result = h3_geodataframe_with_values.h3.h3_to_parent_aggregate(8)
    # TODO: Why does Pandas not preserve the order of groups here?
    index = pd.Index(["881f1d4811fffff", "881f1d4817fffff"], name="h3_08")
    geometry = [Polygon(h3.h3_to_geo_boundary(h, True)) for h in index]
    expected = gpd.GeoDataFrame(
        {"val": [5, 3]}, geometry=geometry, index=index, crs="epsg:4326"
    )

    assert_geodataframe_equal(expected, result)


def test_h3_to_parent_aggregate_no_geometry(h3_dataframe_with_values):
    index = pd.Index(["881f1d4811fffff", "881f1d4817fffff"], name="h3_08")
    expected = pd.DataFrame({"val": [5, 3]}, index=index)
    result = h3_dataframe_with_values.h3.h3_to_parent_aggregate(
        8, return_geometry=False
    )
    pd.testing.assert_frame_equal(expected, result)


def test_h3_k_ring_smoothing_k_vs_weighting(h3_dataframe_with_values):
    result_k = h3_dataframe_with_values.h3.k_ring_smoothing(2)
    result_weighted = h3_dataframe_with_values.h3.k_ring_smoothing(weights=[1, 1, 1])
    pd.testing.assert_frame_equal(result_k, result_weighted)


def test_h3_k_ring_smoothing_0_ring(h3_dataframe_with_values):
    expected = h3_dataframe_with_values.copy().sort_index().astype({"val": float})
    expected.index = expected.index.rename("h3_k_ring")
    result = h3_dataframe_with_values.h3.k_ring_smoothing(0, return_geometry=False)
    pd.testing.assert_frame_equal(expected, result)


def test_h3_k_ring_smoothing_0_ring_weights(h3_dataframe_with_values):
    expected = h3_dataframe_with_values.copy().sort_index().astype({"val": float})
    expected.index = expected.index.rename("h3_k_ring")
    result = h3_dataframe_with_values.h3.k_ring_smoothing(
        weights=[1], return_geometry=False
    )
    pd.testing.assert_frame_equal(expected, result)


def test_h3_k_ring_smoothing_2_ring(h3_dataframe_with_values):
    data = h3_dataframe_with_values.iloc[:1]
    expected = set([1 / 19])
    result = set(data.h3.k_ring_smoothing(2)["val"])
    assert expected == result


def test_h3_k_ring_smoothing_1_ring_weighted(h3_dataframe_with_values):
    data = h3_dataframe_with_values.iloc[:1]
    expected = set([1 / 4, 1 / 8])
    result = set(data.h3.k_ring_smoothing(weights=[2, 1])["val"])
    assert expected == result


def test_polyfill_resample(h3_geodataframe_with_values):
    expected_indices = set().union(
        *[
            {
                "8a1f1d481747fff",
                "8a1f1d48174ffff",
                "8a1f1d481757fff",
                "8a1f1d48175ffff",
                "8a1f1d481767fff",
                "8a1f1d48176ffff",
                "8a1f1d481777fff",
            },
            {
                "8a1f1d481647fff",
                "8a1f1d48164ffff",
                "8a1f1d481657fff",
                "8a1f1d48165ffff",
                "8a1f1d481667fff",
                "8a1f1d48166ffff",
                "8a1f1d481677fff",
            },
            {
                "8a1f1d4810c7fff",
                "8a1f1d4810cffff",
                "8a1f1d4810d7fff",
                "8a1f1d4810dffff",
                "8a1f1d4810e7fff",
                "8a1f1d4810effff",
                "8a1f1d4810f7fff",
            },
        ]
    )
    expected_values = set([1, 2, 5])
    result = h3_geodataframe_with_values.h3.polyfill_resample(10, return_geometry=False)
    assert len(result) == len(h3_geodataframe_with_values) * 7
    assert set(result.index) == expected_indices
    assert set(result["val"]) == expected_values
    assert not result["val"].isna().any()


def test_polyfill_resample_uncovered_rows(basic_geodataframe_polygons):
    basic_geodataframe_polygons.loc[2] = box(0, 0, 3, 3)
    with pytest.warns(UserWarning):
        result = basic_geodataframe_polygons.h3.polyfill_resample(2)

    assert len(result) == 2



from typing import Union, Set, Tuple, List
from shapely.geometry import Polygon, MultiPolygon
from h3 import h3

MultiPolyOrPoly = Union[Polygon, MultiPolygon]


def _extract_coords(polygon: Polygon) -> Tuple[List, List[List]]:
    """Extract the coordinates of outer and inner rings from a Polygon"""
    outer = list(polygon.exterior.coords)
    inners = [list(g.coords) for g in polygon.interiors]
    return outer, inners


def polyfill(
    geometry: MultiPolyOrPoly, resolution: int, geo_json: bool = False
) -> Set[str]:
    """h3.polyfill accepting a shapely (Multi)Polygon

    Parameters
    ----------
    geometry : Polygon or Multipolygon
        Polygon to fill
    resolution : int
        H3 resolution of the filling cells
    geo_json : bool
        If True, coordinates are assumed to be lng/lat. Default: False (lat/lng)

    Returns
    -------
    Set of H3 addresses

    Raises
    ------
    TypeError if geometry is not a Polygon or MultiPolygon
    """
    if isinstance(geometry, Polygon):
        outer, inners = _extract_coords(geometry)
        return h3.polyfill_polygon(outer, resolution, inners, geo_json)

    elif isinstance(geometry, MultiPolygon):
        h3_addresses = []
        for poly in geometry.geoms:
            h3_addresses.extend(polyfill(poly, resolution, geo_json))

        return set(h3_addresses)
    else:
        raise TypeError(f"Unknown type {type(geometry)}")



from typing import Union, Callable, Sequence, Any
import warnings

# Literal is not supported by Python <3.8
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np
import shapely
import pandas as pd
import geopandas as gpd

from h3 import h3
from pandas.core.frame import DataFrame
from geopandas.geodataframe import GeoDataFrame

from .const import COLUMN_H3_POLYFILL
from .util.decorator import catch_invalid_h3_address, doc_standard
from .util.functools import wrapped_partial
from .util.shapely import polyfill

AnyDataFrame = Union[DataFrame, GeoDataFrame]


@pd.api.extensions.register_dataframe_accessor("h3")
class H3Accessor:
    def __init__(self, df: DataFrame):
        self._df = df

    # H3 API
    # These methods simply mirror the H3 API and apply H3 functions to all rows

    def geo_to_h3(
        self,
        resolution: int,
        lat_col: str = "lat",
        lng_col: str = "lng",
        set_index: bool = True,
    ) -> AnyDataFrame:
        """Adds H3 index to (Geo)DataFrame.

        pd.DataFrame: uses `lat_col` and `lng_col` (default `lat` and `lng`)
        gpd.GeoDataFrame: uses `geometry`

        Parameters
        ----------
        resolution : int
            H3 resolution
        lat_col : str
            Name of the latitude column (if used), default 'lat'
        lng_col : str
            Name of the longitude column (if used), default 'lng'
        set_index : bool
            If True, the columns with H3 addresses is set as index, default 'True'

        Returns
        -------
        (Geo)DataFrame with H3 addresses added

        See Also
        --------
        geo_to_h3_aggregate : Extended API method that aggregates points by H3 address

        Examples
        --------
        >>> df = pd.DataFrame({'lat': [50, 51], 'lng':[14, 15]})
        >>> df.h3.geo_to_h3(8)
                         lat  lng
        h3_08
        881e309739fffff   50   14
        881e2659c3fffff   51   15

        >>> df.h3.geo_to_h3(8, set_index=False)
           lat  lng            h3_08
        0   50   14  881e309739fffff
        1   51   15  881e2659c3fffff

        >>> gdf = gpd.GeoDataFrame({'val': [5, 1]},
        >>> geometry=gpd.points_from_xy(x=[14, 15], y=(50, 51)))
        >>> gdf.h3.geo_to_h3(8)
                         val                   geometry
        h3_08
        881e309739fffff    5  POINT (14.00000 50.00000)
        881e2659c3fffff    1  POINT (15.00000 51.00000)

        """
        if isinstance(self._df, gpd.GeoDataFrame):
            lngs = self._df.geometry.x
            lats = self._df.geometry.y
        else:
            lngs = self._df[lng_col]
            lats = self._df[lat_col]

        h3addresses = [
            h3.geo_to_h3(lat, lng, resolution) for lat, lng in zip(lats, lngs)
        ]

        colname = self._format_resolution(resolution)
        assign_arg = {colname: h3addresses}
        df = self._df.assign(**assign_arg)
        if set_index:
            return df.set_index(colname)
        return df

    def h3_to_geo(self) -> GeoDataFrame:
        """Add `geometry` with centroid of each H3 address to the DataFrame.
        Assumes H3 index.

        Returns
        -------
        GeoDataFrame with Point geometry

        Raises
        ------
        ValueError
            When an invalid H3 address is encountered

        See Also
        --------
        h3_to_geo_boundary : Adds a hexagonal cell

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.h3_to_geo()
                         val                   geometry
        881e309739fffff    5  POINT (14.00037 50.00055)
        881e2659c3fffff    1  POINT (14.99715 51.00252)

        """
        return self._apply_index_assign(
            h3.h3_to_geo,
            "geometry",
            lambda x: shapely.geometry.Point(reversed(x)),
            lambda x: gpd.GeoDataFrame(x, crs="epsg:4326"),
        )

    def h3_to_geo_boundary(self) -> GeoDataFrame:
        """Add `geometry` with H3 hexagons to the DataFrame. Assumes H3 index.

        Returns
        -------
        GeoDataFrame with H3 geometry

        Raises
        ------
        ValueError
            When an invalid H3 address is encountered

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.h3_to_geo_boundary()
                         val                                           geometry
        881e309739fffff    5  POLYGON ((13.99527 50.00368, 13.99310 49.99929...
        881e2659c3fffff    1  POLYGON ((14.99201 51.00565, 14.98973 51.00133...
        """
        return self._apply_index_assign(
            wrapped_partial(h3.h3_to_geo_boundary, geo_json=True),
            "geometry",
            lambda x: shapely.geometry.Polygon(x),
            lambda x: gpd.GeoDataFrame(x, crs="epsg:4326"),
        )

    @doc_standard("h3_resolution", "containing the resolution of each H3 address")
    def h3_get_resolution(self) -> AnyDataFrame:
        """
        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.h3_get_resolution()
                         val  h3_resolution
        881e309739fffff    5              8
        881e2659c3fffff    1              8
        """
        return self._apply_index_assign(h3.h3_get_resolution, "h3_resolution")

    @doc_standard("h3_base_cell", "containing the base cell of each H3 address")
    def h3_get_base_cell(self):
        """
        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.h3_get_base_cell()
                         val  h3_base_cell
        881e309739fffff    5            15
        881e2659c3fffff    1            15
        """
        return self._apply_index_assign(h3.h3_get_base_cell, "h3_base_cell")

    @doc_standard("h3_is_valid", "containing the validity of each H3 address")
    def h3_is_valid(self):
        """
        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]}, index=['881e309739fffff', 'INVALID'])
        >>> df.h3.h3_is_valid()
                         val  h3_is_valid
        881e309739fffff    5         True
        INVALID            1        False
        """
        return self._apply_index_assign(h3.h3_is_valid, "h3_is_valid")

    @doc_standard(
        "h3_k_ring", "containing a list H3 addresses within a distance of `k`"
    )
    def k_ring(self, k: int = 1, explode: bool = False) -> AnyDataFrame:
        """
        Parameters
        ----------
        k : int
            the distance from the origin H3 address. Default k = 1
        explode : bool
            If True, will explode the resulting list vertically.
            All other columns' values are copied.
            Default: False

        See Also
        --------
        k_ring_smoothing : Extended API method that distributes numeric values
            to the k-ring cells

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.k_ring(1)
                         val                                          h3_k_ring
        881e309739fffff    5  [881e30973dfffff, 881e309703fffff, 881e309707f...
        881e2659c3fffff    1  [881e2659ddfffff, 881e2659c3fffff, 881e2659cbf...

        >>> df.h3.k_ring(1, explode=True)
                         val        h3_k_ring
        881e2659c3fffff    1  881e2659ddfffff
        881e2659c3fffff    1  881e2659c3fffff
        881e2659c3fffff    1  881e2659cbfffff
        881e2659c3fffff    1  881e2659d5fffff
        881e2659c3fffff    1  881e2659c7fffff
        881e2659c3fffff    1  881e265989fffff
        881e2659c3fffff    1  881e2659c1fffff
        881e309739fffff    5  881e30973dfffff
        881e309739fffff    5  881e309703fffff
        881e309739fffff    5  881e309707fffff
        881e309739fffff    5  881e30973bfffff
        881e309739fffff    5  881e309715fffff
        881e309739fffff    5  881e309739fffff
        881e309739fffff    5  881e309731fffff
        """
        func = wrapped_partial(h3.k_ring, k=k)
        column_name = "h3_k_ring"
        if explode:
            return self.__apply_index_explode(func, column_name, list)
        return self._apply_index_assign(func, column_name, list)

    @doc_standard(
        "h3_hex_ring",
        "containing a list H3 addresses forming a hollow hexagonal ring"
        "at a distance `k`",
    )
    def hex_ring(self, k: int = 1, explode: bool = False) -> AnyDataFrame:
        """
        Parameters
        ----------
        k : int
            the distance from the origin H3 address. Default k = 1
        explode : bool
            If True, will explode the resulting list vertically.
            All other columns' values are copied.
            Default: False

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.hex_ring(1)
                         val                                        h3_hex_ring
        881e309739fffff    5  [881e30973dfffff, 881e309703fffff, 881e309707f...
        881e2659c3fffff    1  [881e2659ddfffff, 881e2659cbfffff, 881e2659d5f...
        >>> df.h3.hex_ring(1, explode=True)
                         val      h3_hex_ring
        881e2659c3fffff    1  881e2659ddfffff
        881e2659c3fffff    1  881e2659cbfffff
        881e2659c3fffff    1  881e2659d5fffff
        881e2659c3fffff    1  881e2659c7fffff
        881e2659c3fffff    1  881e265989fffff
        881e2659c3fffff    1  881e2659c1fffff
        881e309739fffff    5  881e30973dfffff
        881e309739fffff    5  881e309703fffff
        881e309739fffff    5  881e309707fffff
        881e309739fffff    5  881e30973bfffff
        881e309739fffff    5  881e309715fffff
        881e309739fffff    5  881e309731fffff
        """
        func = wrapped_partial(h3.hex_ring, k=k)
        column_name = "h3_hex_ring"
        if explode:
            return self.__apply_index_explode(func, column_name, list)
        return self._apply_index_assign(func, column_name, list)

    @doc_standard("h3_{resolution}", "containing the parent of each H3 address")
    def h3_to_parent(self, resolution: int = None) -> AnyDataFrame:
        """
        Parameters
        ----------
        resolution : int or None
            H3 resolution. If None, then returns the direct parent of each H3 cell.

        See Also
        --------
        h3_to_parent_aggregate : Extended API method that aggregates cells by their
            parent cell

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.h3_to_parent(5)
                         val            h3_05
        881e309739fffff    5  851e3097fffffff
        881e2659c3fffff    1  851e265bfffffff
        """
        # TODO: Test `h3_parent` case
        column = self._format_resolution(resolution) if resolution else "h3_parent"
        return self._apply_index_assign(
            wrapped_partial(h3.h3_to_parent, res=resolution), column
        )

    @doc_standard("h3_center_child", "containing the center child of each H3 address")
    def h3_to_center_child(self, resolution: int = None) -> AnyDataFrame:
        """
        Parameters
        ----------
        resolution : int or None
            H3 resolution. If none, then returns the child of resolution
            directly below that of each H3 cell

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                    index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.h3_to_center_child()
                         val  h3_center_child
        881e309739fffff    5  891e3097383ffff
        881e2659c3fffff    1  891e2659c23ffff
        """
        return self._apply_index_assign(
            wrapped_partial(h3.h3_to_center_child, res=resolution), "h3_center_child"
        )

    @doc_standard(
        COLUMN_H3_POLYFILL,
        "containing a list H3 addresses whose centroid falls into the Polygon",
    )
    def polyfill(self, resolution: int, explode: bool = False) -> AnyDataFrame:
        """
        Parameters
        ----------
        resolution : int
            H3 resolution
        explode : bool
            If True, will explode the resulting list vertically.
            All other columns' values are copied.
            Default: False

        See Also
        --------
        polyfill_resample : Extended API method that distributes the polygon's values
            to the H3 cells contained in it

        Examples
        --------
        >>> from shapely.geometry import box
        >>> gdf = gpd.GeoDataFrame(geometry=[box(0, 0, 1, 1)])
        >>> gdf.h3.polyfill(4)
                                                    geometry                                        h3_polyfill
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  [84754e3ffffffff, 84754c7ffffffff, 84754c5ffff...  # noqa E501
        >>> gdf.h3.polyfill(4, explode=True)
                                                    geometry      h3_polyfill
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  84754e3ffffffff
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  84754c7ffffffff
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  84754c5ffffffff
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  84754ebffffffff
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  84754edffffffff
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  84754e1ffffffff
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  84754e9ffffffff
        0  POLYGON ((1.00000 0.00000, 1.00000 1.00000, 0....  8475413ffffffff
        """

        def func(row):
            return list(polyfill(row.geometry, resolution, True))

        result = self._df.apply(func, axis=1)

        if not explode:
            assign_args = {COLUMN_H3_POLYFILL: result}
            return self._df.assign(**assign_args)

        result = result.explode().to_frame(COLUMN_H3_POLYFILL)

        return self._df.join(result)

    @doc_standard("h3_cell_area", "containing the area of each H3 address")
    def cell_area(
        self, unit: Literal["km^2", "m^2", "rads^2"] = "km^2"
    ) -> AnyDataFrame:
        """
        Parameters
        ----------
        unit : str, options: 'km^2', 'm^2', or 'rads^2'
            Unit for area result. Default: 'km^2`

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.cell_area()
                         val  h3_cell_area
        881e309739fffff    5      0.695651
        881e2659c3fffff    1      0.684242
        """
        return self._apply_index_assign(
            wrapped_partial(h3.cell_area, unit=unit), "h3_cell_area"
        )

    # H3-Pandas Extended API
    # These methods extend the API to provide a convenient way to simplify workflows

    def geo_to_h3_aggregate(
        self,
        resolution: int,
        operation: Union[dict, str, Callable] = "sum",
        lat_col: str = "lat",
        lng_col: str = "lng",
        return_geometry: bool = True,
    ) -> DataFrame:
        """Adds H3 index to DataFrame, groups points with the same index
        and performs `operation`.

        pd.DataFrame: uses `lat_col` and `lng_col` (default `lat` and `lng`)
        gpd.GeoDataFrame: uses `geometry`

        Parameters
        ----------
        resolution : int
            H3 resolution
        operation : Union[dict, str, Callable]
            Argument passed to DataFrame's `agg` method, default 'sum'
        lat_col : str
            Name of the latitude column (if used), default 'lat'
        lng_col : str
            Name of the longitude column (if used), default 'lng'
        return_geometry: bool
            (Optional) Whether to add a `geometry` column with the hexagonal cells.
            Default = True

        Returns
        -------
        (Geo)DataFrame aggregated by H3 address into which each row's point falls

        See Also
        --------
        geo_to_h3 : H3 API method upon which this function builds

        Examples
        --------
        >>> df = pd.DataFrame({'lat': [50, 51], 'lng':[14, 15], 'val': [10, 1]})
        >>> df.h3.geo_to_h3(1)
                         lat  lng  val
        h3_01
        811e3ffffffffff   50   14   10
        811e3ffffffffff   51   15    1
        >>> df.h3.geo_to_h3_aggregate(1)
                         val                                           geometry
        h3_01
        811e3ffffffffff   11  POLYGON ((12.34575 50.55428, 12.67732 46.40696...
        >>> df = pd.DataFrame({'lat': [50, 51], 'lng':[14, 15], 'val': [10, 1]})
        >>> df.h3.geo_to_h3_aggregate(1, operation='mean')
                         val                                           geometry
        h3_01
        811e3ffffffffff  5.5  POLYGON ((12.34575 50.55428, 12.67732 46.40696...
        >>> df.h3.geo_to_h3_aggregate(1, return_geometry=False)
                         val
        h3_01
        811e3ffffffffff   11
        """
        grouped = pd.DataFrame(
            self.geo_to_h3(resolution, lat_col, lng_col, False)
            .drop(columns=[lat_col, lng_col, "geometry"], errors="ignore")
            .groupby(self._format_resolution(resolution))
            .agg(operation)
        )
        return grouped.h3.h3_to_geo_boundary() if return_geometry else grouped

    def h3_to_parent_aggregate(
        self,
        resolution: int,
        operation: Union[dict, str, Callable] = "sum",
        return_geometry: bool = True,
    ) -> GeoDataFrame:
        """Assigns parent cell to each row, groups by it and performs `operation`.
        Assumes H3 index.

        Parameters
        ----------
        resolution : int
            H3 resolution
        operation : Union[dict, str, Callable]
            Argument passed to DataFrame's `agg` method, default 'sum'
        return_geometry: bool
            (Optional) Whether to add a `geometry` column with the hexagonal cells.
            Default = True

        Returns
        -------
        (Geo)DataFrame aggregated by the parent of each H3 address

        Raises
        ------
        ValueError
            When an invalid H3 address is encountered

        See Also
        --------
        h3_to_parent : H3 API method upon which this function builds

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.h3_to_parent(1)
                         val            h3_01
        881e309739fffff    5  811e3ffffffffff
        881e2659c3fffff    1  811e3ffffffffff
        >>> df.h3.h3_to_parent_aggregate(1)
                         val                                           geometry
        h3_01
        811e3ffffffffff    6  POLYGON ((12.34575 50.55428, 12.67732 46.40696...
        >>> df.h3.h3_to_parent_aggregate(1, operation='mean')
                         val                                           geometry
        h3_01
        811e3ffffffffff    3  POLYGON ((12.34575 50.55428, 12.67732 46.40696...
        >>> df.h3.h3_to_parent_aggregate(1, return_geometry=False)
                         val
        h3_01
        811e3ffffffffff    6
        """
        parent_h3addresses = [
            catch_invalid_h3_address(h3.h3_to_parent)(h3address, resolution)
            for h3address in self._df.index
        ]
        h3_parent_column = self._format_resolution(resolution)
        kwargs_assign = {h3_parent_column: parent_h3addresses}
        grouped = (
            self._df.assign(**kwargs_assign)
            .groupby(h3_parent_column)[[c for c in self._df.columns if c != "geometry"]]
            .agg(operation)
        )

        return grouped.h3.h3_to_geo_boundary() if return_geometry else grouped

    # TODO: Needs to allow for handling relative values (e.g. percentage)
    # TODO: Will possibly fail in many cases (what are the existing columns?)
    # TODO: New cell behaviour
    def k_ring_smoothing(
        self,
        k: int = None,
        weights: Sequence[float] = None,
        return_geometry: bool = True,
    ) -> AnyDataFrame:
        """Experimental. Creates a k-ring around each input cell and distributes
        the cell's values.

        The values are distributed either
         - uniformly (by setting `k`) or
         - by weighing their values using `weights`.

        Only numeric columns are modified.

        Parameters
        ----------
        k : int
            The distance from the origin H3 address
        weights : Sequence[float]
            Weighting of the values based on the distance from the origin.
            First weight corresponds to the origin.
            Values are be normalized to add up to 1.
        return_geometry: bool
            (Optional) Whether to add a `geometry` column with the hexagonal cells.
            Default = True

        Returns
        -------
        (Geo)DataFrame with smoothed values

        See Also
        --------
        k_ring : H3 API method upon which this method builds

        Examples
        --------
        >>> df = pd.DataFrame({'val': [5, 1]},
        >>>                   index=['881e309739fffff', '881e2659c3fffff'])
        >>> df.h3.k_ring_smoothing(1)
                              val                                           geometry
        h3_k_ring
        881e265989fffff  0.142857  POLYGON ((14.99488 50.99821, 14.99260 50.99389...
        881e2659c1fffff  0.142857  POLYGON ((14.97944 51.00758, 14.97717 51.00326...
        881e2659c3fffff  0.142857  POLYGON ((14.99201 51.00565, 14.98973 51.00133...
        881e2659c7fffff  0.142857  POLYGON ((14.98231 51.00014, 14.98004 50.99582...
        881e2659cbfffff  0.142857  POLYGON ((14.98914 51.01308, 14.98687 51.00877...
        881e2659d5fffff  0.142857  POLYGON ((15.00458 51.00371, 15.00230 50.99940...
        881e2659ddfffff  0.142857  POLYGON ((15.00171 51.01115, 14.99943 51.00684...
        881e309703fffff  0.714286  POLYGON ((13.99235 50.01119, 13.99017 50.00681...
        881e309707fffff  0.714286  POLYGON ((13.98290 50.00555, 13.98072 50.00116...
        881e309715fffff  0.714286  POLYGON ((14.00473 50.00932, 14.00255 50.00494...
        881e309731fffff  0.714286  POLYGON ((13.99819 49.99617, 13.99602 49.99178...
        881e309739fffff  0.714286  POLYGON ((13.99527 50.00368, 13.99310 49.99929...
        881e30973bfffff  0.714286  POLYGON ((14.00765 50.00181, 14.00547 49.99742...
        881e30973dfffff  0.714286  POLYGON ((13.98582 49.99803, 13.98364 49.99365...
        >>> df.h3.k_ring_smoothing(weights=[2, 1])
                           val                                           geometry
        h3_hex_ring
        881e265989fffff  0.125  POLYGON ((14.99488 50.99821, 14.99260 50.99389...
        881e2659c1fffff  0.125  POLYGON ((14.97944 51.00758, 14.97717 51.00326...
        881e2659c3fffff  0.250  POLYGON ((14.99201 51.00565, 14.98973 51.00133...
        881e2659c7fffff  0.125  POLYGON ((14.98231 51.00014, 14.98004 50.99582...
        881e2659cbfffff  0.125  POLYGON ((14.98914 51.01308, 14.98687 51.00877...
        881e2659d5fffff  0.125  POLYGON ((15.00458 51.00371, 15.00230 50.99940...
        881e2659ddfffff  0.125  POLYGON ((15.00171 51.01115, 14.99943 51.00684...
        881e309703fffff  0.625  POLYGON ((13.99235 50.01119, 13.99017 50.00681...
        881e309707fffff  0.625  POLYGON ((13.98290 50.00555, 13.98072 50.00116...
        881e309715fffff  0.625  POLYGON ((14.00473 50.00932, 14.00255 50.00494...
        881e309731fffff  0.625  POLYGON ((13.99819 49.99617, 13.99602 49.99178...
        881e309739fffff  1.250  POLYGON ((13.99527 50.00368, 13.99310 49.99929...
        881e30973bfffff  0.625  POLYGON ((14.00765 50.00181, 14.00547 49.99742...
        881e30973dfffff  0.625  POLYGON ((13.98582 49.99803, 13.98364 49.99365...
        >>> df.h3.k_ring_smoothing(1, return_geometry=False)
                              val
        h3_k_ring
        881e265989fffff  0.142857
        881e2659c1fffff  0.142857
        881e2659c3fffff  0.142857
        881e2659c7fffff  0.142857
        881e2659cbfffff  0.142857
        881e2659d5fffff  0.142857
        881e2659ddfffff  0.142857
        881e309703fffff  0.714286
        881e309707fffff  0.714286
        881e309715fffff  0.714286
        881e309731fffff  0.714286
        881e309739fffff  0.714286
        881e30973bfffff  0.714286
        881e30973dfffff  0.714286
        """
        if sum([weights is None, k is None]) != 1:
            raise ValueError("Exactly one of `k` and `weights` must be set.")

        # If weights are all equal, use the computationally simpler option
        if (weights is not None) and (len(set(weights)) == 1):
            k = len(weights) - 1
            weights = None

        # Unweighted case
        if weights is None:
            result = (
                self._df.h3.k_ring(k, explode=True)
                .groupby("h3_k_ring")
                .sum()
                .divide((1 + 3 * k * (k + 1)))
            )

            return result.h3.h3_to_geo_boundary() if return_geometry else result

        if len(weights) == 0:
            raise ValueError("Weights cannot be empty.")

        # Weighted case
        weights = np.array(weights)
        multipliers = np.array([1] + [i * 6 for i in range(1, len(weights))])
        weights = weights / (weights * multipliers).sum()

        # This should be exploded hex ring
        def weighted_hex_ring(df, k, normalized_weight):
            return df.h3.hex_ring(k, explode=True).h3._multiply_numeric(
                normalized_weight
            )

        result = (
            pd.concat(
                [
                    weighted_hex_ring(self._df, i, weights[i])
                    for i in range(len(weights))
                ]
            )
            .groupby("h3_hex_ring")
            .sum()
        )

        return result.h3.h3_to_geo_boundary() if return_geometry else result

    def polyfill_resample(
        self, resolution: int, return_geometry: bool = True
    ) -> AnyDataFrame:
        """Experimental. Currently essentially polyfill(..., explode=True) that
        sets the H3 index and adds the H3 cell geometry.

        Parameters
        ----------
        resolution : int
            H3 resolution
        return_geometry: bool
            (Optional) Whether to add a `geometry` column with the hexagonal cells.
            Default = True

        Returns
        -------
        (Geo)DataFrame with H3 cells with centroids within the input polygons.

        See Also
        --------
        polyfill : H3 API method upon which this method builds

        Examples
        --------
        >>> from shapely.geometry import box
        >>> gdf = gpd.GeoDataFrame(geometry=[box(0, 0, 1, 1)])
        >>> gdf.h3.polyfill_resample(4)
                         index                                           geometry
        h3_polyfill
        84754e3ffffffff      0  POLYGON ((0.33404 -0.11975, 0.42911 0.07901, 0...
        84754c7ffffffff      0  POLYGON ((0.92140 -0.03115, 1.01693 0.16862, 0...
        84754c5ffffffff      0  POLYGON ((0.91569 0.33807, 1.01106 0.53747, 0....
        84754ebffffffff      0  POLYGON ((0.62438 0.10878, 0.71960 0.30787, 0....
        84754edffffffff      0  POLYGON ((0.32478 0.61394, 0.41951 0.81195, 0....
        84754e1ffffffff      0  POLYGON ((0.32940 0.24775, 0.42430 0.44615, 0....
        84754e9ffffffff      0  POLYGON ((0.61922 0.47649, 0.71427 0.67520, 0....
        8475413ffffffff      0  POLYGON ((0.91001 0.70597, 1.00521 0.90497, 0....
        """
        result = self._df.h3.polyfill(resolution, explode=True)
        uncovered_rows = result[COLUMN_H3_POLYFILL].isna()
        n_uncovered_rows = uncovered_rows.sum()
        if n_uncovered_rows > 0:
            warnings.warn(
                f"{n_uncovered_rows} rows did not generate a H3 cell."
                "Consider using a finer resolution."
            )
            result = result.loc[~uncovered_rows]

        result = result.reset_index().set_index(COLUMN_H3_POLYFILL)

        return result.h3.h3_to_geo_boundary() if return_geometry else result

    # Private methods

    def _apply_index_assign(
        self,
        func: Callable,
        column_name: str,
        processor: Callable = lambda x: x,
        finalizer: Callable = lambda x: x,
    ) -> Any:
        """Helper method. Applies `func` to index and assigns the result to `column`.

        Parameters
        ----------
        func : Callable
            single-argument function to be applied to each H3 address
        column_name : str
            name of the resulting column
        processor : Callable
            (Optional) further processes the result of func. Default: identity
        finalizer : Callable
            (Optional) further processes the resulting dataframe. Default: identity

        Returns
        -------
        Dataframe with column `column` containing the result of `func`.
        If using `finalizer`, can return anything the `finalizer` returns.
        """
        func = catch_invalid_h3_address(func)
        result = [processor(func(h3address)) for h3address in self._df.index]
        assign_args = {column_name: result}
        return finalizer(self._df.assign(**assign_args))

    def __apply_index_explode(
        self,
        func: Callable,
        column_name: str,
        processor: Callable = lambda x: x,
        finalizer: Callable = lambda x: x,
    ) -> Any:
        """Helper method. Applies a list-making `func` to index and performs
        a vertical explode.
        Any additional values are simply copied to all the rows.

        Parameters
        ----------
        func : Callable
            single-argument function to be applied to each H3 address
        column_name : str
            name of the resulting column
        processor : Callable
            (Optional) further processes the result of func. Default: identity
        finalizer : Callable
            (Optional) further processes the resulting dataframe. Default: identity

        Returns
        -------
        Dataframe with column `column` containing the result of `func`.
        If using `finalizer`, can return anything the `finalizer` returns.
        """
        func = catch_invalid_h3_address(func)
        result = (
            pd.DataFrame.from_dict(
                {h3address: processor(func(h3address)) for h3address in self._df.index},
                orient="index",
            )
            .stack()
            .to_frame(column_name)
            .reset_index(level=1, drop=True)
        )
        result = self._df.join(result)
        return finalizer(result)

    # TODO: types, doc, ..
    def _multiply_numeric(self, value):
        columns_numeric = self._df.select_dtypes(include=["number"]).columns
        assign_args = {
            column: self._df[column].multiply(value) for column in columns_numeric
        }
        return self._df.assign(**assign_args)

    @staticmethod
    def _format_resolution(resolution: int) -> str:
        return f"h3_{str(resolution).zfill(2)}"



from functools import wraps
from typing import Callable
from h3 import H3CellError


def catch_invalid_h3_address(f: Callable) -> Callable:
    """Wrapper that catches potential invalid H3 addresses.

    Parameters
    ----------
    f : Callable

    Returns
    -------
    The return value of f, or a ValueError if f threw ValueError, TypeError,
    or H3CellError

    Raises
    ------
    ValueError
        When an invalid H3 address is encountered
    """

    @wraps(f)
    def safe_f(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except (TypeError, ValueError, H3CellError) as e:
            message = "H3 method raised an error. Is the H3 address correct?"
            message += f"\nCaller: {f.__name__}({_print_signature(*args, **kwargs)})"
            message += f"\nOriginal error: {repr(e)}"
            raise ValueError(message)

    return safe_f


# TODO: Test
def doc_standard(column_name: str, description: str) -> Callable:
    """Wrapper to provide a standard apply-to-H3-index docstring"""

    def doc_decorator(f):
        @wraps(f)
        def doc_f(*args, **kwargs):
            return f(*args, **kwargs)

        parameters = f.__doc__ or ""

        doc = f"""Adds the column `{column_name}` {description}. Assumes H3 index.
        {parameters}
        Returns
        -------
        Geo(DataFrame) with `{column_name}` column added

        Raises
        ------
        ValueError
            When an invalid H3 address is encountered
        """

        doc_f.__doc__ = doc
        return doc_f

    return doc_decorator


def _print_signature(*args, **kwargs):
    signature = []
    if args:
        signature.append(", ".join([repr(a) for a in args]))
    if kwargs:
        signature.append(", ".join({f"{repr(k)}={repr(v)}" for k, v in kwargs.items()}))

    return ", ".join(signature)



from setuptools import setup, find_packages
from os import path
import versioneer

# read the contents of README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

version = versioneer.get_version()
download_url = f"https://github.com/DahnJ/H3-Pandas/archive/refs/tags/{version}.tar.gz"

setup(
    name="h3pandas",
    version=version,
    cmdclass=versioneer.get_cmdclass(),
    license="MIT",
    description="Integration of H3 and GeoPandas",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Dahn",
    author_email="dahnjahn@gmail.com",
    url="https://github.com/DahnJ/H3-Pandas",
    download_url=download_url,
    keywords=[
        "python",
        "h3",
        "geospatial",
        "geopandas",
        "pandas",
        "integration",
        "hexagons-are-bestagons",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    packages=find_packages(),
    setup_requires=[],
    install_requires=[
        "geopandas",
        "numpy",
        "pandas",
        "shapely",
        "h3",
        "numpy",
        "typing-extensions",
    ],
    python_requires=">=3.6",
    extras_require={
        "test": ["pytest", "pytest-cov", "flake8"],
        "docs": ["sphinx", "numpydoc", "pytest-sphinx-theme", "typing-extensions"],
    },
)
